/* PSPP - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

/* AIX requires this to be the first thing in the file.  */
#include <config.h>
#if __GNUC__
#define alloca __builtin_alloca
#else
#if HAVE_ALLOCA_H
#include <alloca.h>
#else
#ifdef _AIX
#pragma alloca
#else
#ifndef alloca			/* predefined by HP cc +Olibcalls */
char *alloca ();
#endif
#endif
#endif
#endif

#include <ctype.h>
#include <assert.h>
#include <stdarg.h>
#include <limits.h>
#include "error.h"
#include "str.h"
#include "misc.h"
#include "arena.h"
#include "output.h"
#include "som.h"
#include "tab.h"
#include "var.h"
#include "misc.h"
#include "command.h"

#undef DEBUGGING
/*#define DEBUGGING 1 */
#include "debug-print.h"

/* Creates a table with NC columns and NR rows.  If RESIZABLE is
   nonzero then the table's size can be changed later; otherwise, its
   size is fixed. */
tab_table *
tab_create (int nc, int nr, int resizable)
{
  void *(*alloc_func) (arena **, size_t);

  tab_table *t;
  
  {
    arena *container;
    
    arena_create (&container);
    t = arena_alloc (&container, sizeof (tab_table));
    t->container = container;
  }
  
  t->col_style = SOM_COL_NONE;
  t->col_group = 0;
  t->title.l = 0;
  t->title.s = NULL;
  t->flags = SOMF_NONE;
  t->nr = nr;
  t->nc = nc;
  t->l = t->r = t->t = t->b = 0;

  alloc_func = resizable ? arena_malloc : arena_alloc;

  t->cc = alloc_func (&t->container, nr * nc * sizeof (a_string));
  t->ct = alloc_func (&t->container, nr * nc);
  t->rh = alloc_func (&t->container, nc * (nr + 1));
  memset (t->rh, 0, nc * (nr + 1));
  t->hrh = alloc_func (&t->container, sizeof (int) * (nr + 1));
  memset (t->hrh, 0, sizeof (int) * (nr + 1));
  t->trh = alloc_func (&t->container, sizeof (int) * (nr + 1));
  memset (t->trh, 0, sizeof (int) * (nr + 1));
  t->rv = alloc_func (&t->container, (nc + 1) * nr);
  memset (t->rv, 0, (nc + 1) * nr);
  t->wrv = alloc_func (&t->container, sizeof (int) * (nc + 1));
  memset (t->wrv, 0, sizeof (int) * (nc + 1));
  t->trv = alloc_func (&t->container, sizeof (int) * (nc + 1));
  memset (t->trv, 0, sizeof (int) * (nc + 1));
  t->dim = NULL;
  t->w = alloc_func (&t->container, nc * sizeof (som_dimension));
  t->h = alloc_func (&t->container, nr * sizeof (som_dimension));
  t->max_stack_height = 0;
  
#if GLOBAL_DEBUGGING
  {
    int i;
    a_string *c;

    for (i = nc * nr, c = t->cc; i--; c++)
      {
	c->l = -1;
	c->s = NULL;
      }
  }
#endif

  return t;
}

/* Destroys table T. */
void
tab_destroy (tab_table *t)
{
  arena_destroy (&t->container);
}

/* Sets the size of a table, in rows.  Use only to reduce the number
   of rows in a table, since it does not change the amount of
   allocated memory. */
void
tab_height (tab_table *t, int nr)
{
  t->nr = nr;
}

/* Changes either or both dimensions of a table.  Consider using the
   above routine instead if it won't waste a lot of space.

   Changing the number of columns in a table is particularly expensive
   in space and time.  Avoid doing such.  FIXME: In fact, transferring
   of rules isn't even implemented yet. */
void
tab_realloc (tab_table *t, int nc, int nr)
{
  assert (nc == t->nc);
  
  if (nc != t->nc)
    {
      int mr1 = min (nr, t->nr);
      int mc1 = min (nc, t->nc);
      
      a_string *new_cc;
      unsigned char *new_ct;
      int r;

      new_cc = arena_malloc (&t->container, nr * nc * sizeof (a_string));
      new_ct = arena_malloc (&t->container, nr * nc);
      for (r = 0; r < mr1; r++)
	{
	  memcpy (&new_cc[r * nc], &t->cc[r * t->nc], mc1 * sizeof (a_string));
	  memcpy (&new_ct[r * nc], &t->ct[r * t->nc], mc1);
#if DEBUGGING
	  if (nc > t->nc)
	    {
	      a_string *c;

	      for (c = &new_cc[r * nc + t->nc]; c < &new_cc[(r + 1) * nc]; c++)
		{
		  c->l = 0;
		  c->s = NULL;
		}
	      memset (&new_ct[r * nc + t->nc], 0, nc - t->nc);
	    }
#endif
	}
      arena_free (t->cc);
      arena_free (t->ct);
      t->cc = new_cc;
      t->ct = new_ct;
    }
  else if (nr != t->nr)
    {
      t->cc = arena_realloc (&t->container, t->cc,
			     nr * nc * sizeof (a_string));
      t->ct = arena_realloc (&t->container, t->ct, nr * nc);

      t->rh = arena_realloc (&t->container, t->rh, nc * (nr + 1));
      t->rv = arena_realloc (&t->container, t->rv, (nc + 1) * nr);
      t->trh = arena_realloc (&t->container, t->trh, sizeof (int) * (nr + 1));
      t->hrh = arena_realloc (&t->container, t->hrh, sizeof (int) * (nr + 1));
      t->h = arena_realloc (&t->container, t->h, sizeof (som_dimension) * nr);
      
      if (nr > t->nr)
	{
	  memset (&t->rh[nc * (t->nr + 1)], 0, (nr - t->nr) * nc);
	  memset (&t->rv[(nc + 1) * t->nr], 0, (nr - t->nr) * (nc + 1));
	  memset (&t->trh[t->nr + 1], 0, sizeof (int) * (nr - t->nr));
	}
    }
  else
    return;

#if GLOBAL_DEBUGGING
  if (nr > t->nr)
    {
      a_string *c;

      for (c = &t->cc[t->nr * nc]; c < &t->cc[nr * nc]; c++)
	{
	  c->l = -1;
	  c->s = NULL;
	}
      memset (&t->ct[nc * t->nr], 0, nc * (nr - t->nr));
    }
#endif
  
  t->nr = nr;
  t->nc = nc;
}

/* Sets the number of header rows on each side of TABLE to L on the
   left, R on the right, T on the top, B on the bottom.  Header rows
   are repeated when a table is broken across multiple columns or
   multiple pages. */
void
tab_headers (tab_table *table, int l, int r, int t, int b)
{
  table->l = l;
  table->r = r;
  table->t = t;
  table->b = b;
}

/* Set up table T so that, when it is an appropriate size, it will be
   displayed across the page in columns.

   STYLE is a SOM_COL_* constant.  GROUP is the number of rows to take
   as a unit. */
void
tab_columns (tab_table *t, int style, int group)
{
  t->col_style = style;
  t->col_group = group;
}

/* Rules. */

/* Draws a vertical line to the left of cells at horizontal position X
   from Y1 to Y2 inclusive in style STYLE, if style is not -1. */
void
tab_vline (tab_table *t, int style, int x, int y1, int y2)
{
  int y;

#if GLOBAL_DEBUGGING
  if (x < 0 || x > t->nc || y1 < 0 || y1 >= t->nr || y2 < 0 || y2 >= t->nr)
    {
      printf (_("bad vline: x=%d y=(%d,%d) in table size (%d,%d)\n"),
	      x, y1, y2, t->nc, t->nr);
      return;
    }
#endif
  if (style != -1)
    {
      if ((style & TAL_SPACING) == 0)
	for (y = y1; y <= y2; y++)
	  t->rv[x + (t->nc + 1) * y] = style;
      t->trv[x] |= (1 << (style & ~TAL_SPACING));
    }
}

/* Draws a horizontal line above cells at vertical position Y from X1
   to X2 inclusive in style STYLE, if style is not -1. */
void
tab_hline (tab_table * t, int style, int x1, int x2, int y)
{
  int x;

#if GLOBAL_DEBUGGING
  if (x1 < 0 || x1 >= t->nc || x2 < 0 || x2 >= t->nc || y < 0 || y > t->nr)
    {
      printf (_("bad hline: x=(%d,%d) y=%d in table size (%d,%d)\n"),
	      x1, x2, y, t->nc, t->nr);
      return;
    }
#endif
  if (style != -1)
    {
      if ((style & TAL_SPACING) == 0)
	for (x = x1; x <= x2; x++)
	  t->rh[x + t->nc * y] = style;
      t->trh[y] |= (1 << (style & ~TAL_SPACING));
    }
}

/* Draws a box around cells (X1,Y1)-(X2,Y2) inclusive with horizontal
   lines of style F_H and vertical lines of style F_V.  Fills the
   interior of the box with horizontal lines of style I_H and vertical
   lines of style I_V.  Any of the line styles may be -1 to avoid
   drawing those lines.  This is distinct from 0, which draws a null
   line. */
void
tab_box (tab_table *t, int f_h, int f_v, int i_h, int i_v,
	 int x1, int y1, int x2, int y2)
{
#if GLOBAL_DEBUGGING
  if (x1 < 0 || x1 >= t->nc || x2 < 0 || x2 >= t->nc
      || y1 < 0 || y1 >= t->nr || y2 < 0 || y2 >= t->nr)
    {
      printf (_("bad box: (%d,%d)-(%d,%d) in table size (%d,%d)\n"),
	      x1, y1, x2, y2, t->nc, t->nr);
      return;
    }
#endif

  if (f_h != -1)
    {
      int x;
      if ((f_h & TAL_SPACING) == 0)
	for (x = x1; x <= x2; x++)
	  {
	    t->rh[x + t->nc * y1] = f_h;
	    t->rh[x + t->nc * (y2 + 1)] = f_h;
	  }
      t->trh[y1] |= (1 << (f_h & ~TAL_SPACING));
      t->trh[y2 + 1] |= (1 << (f_h & ~TAL_SPACING));
    }
  if (f_v != -1)
    {
      int y;
      if ((f_v & TAL_SPACING) == 0)
	for (y = y1; y <= y2; y++)
	  {
	    t->rv[x1 + (t->nc + 1) * y] = f_v;
	    t->rv[(x2 + 1) + (t->nc + 1) * y] = f_v;
	  }
      t->trv[x1] |= (1 << (f_v & ~TAL_SPACING));
      t->trv[x2 + 1] |= (1 << (f_v & ~TAL_SPACING));
    }

  if (i_h != -1)
    {
      int y;
      
      for (y = y1 + 1; y <= y2; y++)
	{
	  int x;

	  if ((i_h & TAL_SPACING) == 0)
	    for (x = x1; x <= x2; x++)
	      t->rh[x + t->nc * y] = i_h;

	  t->trh[y] |= (1 << (i_h & ~TAL_SPACING));
	}
    }
  if (i_v != -1)
    {
      int x;
      
      for (x = x1 + 1; x <= x2; x++)
	{
	  int y;
	  
	  if ((i_v & TAL_SPACING) == 0)
	    for (y = y1; y <= y2; y++)
	      t->rv[x + (t->nc + 1) * y] = i_v;

	  t->trv[x] |= (1 << (i_v & ~TAL_SPACING));
	}
    }
}

/* Formats text TEXT and arguments ARGS as indicated in OPT and sets
   the resultant string into S in TABLE's arena. */
static void
text_format (tab_table *table, int opt, const char *text, va_list args,
	     a_string *s)
{
  int len;
  
  assert (table && text);
  
  if (opt & TAT_PRINTF)
    {
      char *temp_buf = local_alloc (1024);
      
      len = nvsprintf (temp_buf, text, args);
      text = temp_buf;
    }
  else
    len = strlen (text);
  
  if ((opt & TAT_TITLE) == TAT_TITLE)
    {
      s->s = arena_alloc (&table->container, len + 5);
      s->l = len + 5;

      s->s[0] = 0;
      STORE_2 (&s->s[1], 5);
      s->s[3] = TAG_FONT_BY_POSITION;
      s->s[4] = OUTP_F_BI;
      memcpy (&s->s[5], text, len);
    }
  else
    {
      s->s = arena_alloc (&table->container, len);
      s->l = len;
      memcpy (s->s, text, len);
    }
  
  if (opt & TAT_PRINTF)
    local_free (text);
}

/* Set the title of table T to TITLE, which is formatted with printf
   if FORMAT is nonzero. */
void
tab_title (tab_table *t, int format, const char *title, ...)
{
  va_list args;

  va_start (args, title);
  text_format (t, format ? TAT_PRINTF : TAT_NONE, title, args, &t->title);
  va_end (args);
}


/* Cells. */

/* Sets cell (C,R) in TABLE as an empty cell. */
void
tab_null (tab_table * table, int c, int r)
{
#if GLOBAL_DEBUGGING
  if (c < 0 || r < 0 || c >= table->nc || r >= table->nr)
    {
      printf (_("tab_null(): bad cell (%d,%d) in table size (%d,%d)\n"),
	      c, r, table->nc, table->nr);
      return;
    }
#endif

  table->ct[c + r * table->nc] = TAB_NONE;
  table->cc[c + r * table->nc].l = 0;
  table->cc[c + r * table->nc].s = NULL;
}

/* Sets cell (C,R) in TABLE, with options OPT, to have a value taken
   from a vector at V, displayed with format spec F. */
void
tab_value (tab_table *table, int c, int r, unsigned char opt,
	   const value *v, const fmt_spec *f)
{
  char *contents;
  value temp_val;

  assert (table && v && f);
#if GLOBAL_DEBUGGING
  if (c < 0 || r < 0 || c >= table->nc || r >= table->nr)
    {
      printf (_("tab_value(): bad cell (%d,%d) in table size (%d,%d)\n"),
	      c, r, table->nc, table->nr);
      return;
    }
#endif

  contents = arena_alloc (&table->container, f->w);
  table->cc[c + r * table->nc].l = f->w;
  table->cc[c + r * table->nc].s = contents;
  table->ct[c + r * table->nc] = opt;
  
  if (formats[f->type].cat & FCAT_STRING)
    {
      temp_val.c = (char *) v->s;
      v = &temp_val;
    }
  convert_format_to_string (contents, f, v);
}

/* Sets cell (C,R) in TABLE, with options OPT, to have value VAL
   with NDEC decimal places. */
void
tab_float (tab_table *table, int c, int r, unsigned char opt,
	   double val, int w, int d)
{
  char *contents;
  char buf[40], *cp;
  
  fmt_spec f;

  assert (w <= 40);
  
  f.type = FMT_F;
  f.w = w;
  f.d = d;
  
  assert (table);
#if GLOBAL_DEBUGGING
  if (c < 0 || r < 0 || c >= table->nc || r >= table->nr)
    {
      printf (_("tab_float(): bad cell (%d,%d) in table size (%d,%d)\n"),
	      c, r, table->nc, table->nr);
      return;
    }
#endif

  convert_format_to_string (buf, &f, (value *) &val);
  cp = buf;
  while (isspace ((unsigned char) *cp) && cp < &buf[w])
    cp++;
  f.w = w - (cp - buf);

  contents = arena_alloc (&table->container, f.w);
  table->cc[c + r * table->nc].l = f.w;
  table->cc[c + r * table->nc].s = contents;
  table->ct[c + r * table->nc] = opt;
  memcpy (contents, cp, f.w);
}

/* Sets cell (C,R) in TABLE, with options OPT, to have text value
   TEXT. */
void
tab_text (tab_table *table, int c, int r, unsigned opt, const char *text, ...)
{
  va_list args;

  assert (table && text);
#if GLOBAL_DEBUGGING
  if (c < 0 || r < 0 || c >= table->nc || r >= table->nr)
    {
      printf (_("tab_text(): bad cell (%d,%d) in table size (%d,%d)\n"),
	      c, r, table->nc, table->nr);
      return;
    }
#endif
    
  va_start (args, text);
  text_format (table, opt, text, args, &table->cc[c + r * table->nc]);
  table->ct[c + r * table->nc] = opt;
  va_end (args);
}

/* Joins cells (X1,X2)-(Y1,Y2) inclusive in TABLE, and sets them with
   options OPT to have text value TEXT. */
void
tab_joint_text (tab_table *table, int x1, int y1, int x2, int y2,
		unsigned opt, const char *text, ...)
{
  tab_joined_cell *j;

  assert (table && text
	  && x1 >= 0 && x1 < table->nc && x2 >= x1 && x2 < table->nc
	  && y1 >= 0 && y1 < table->nr && y2 >= y1 && y2 < table->nr);
  
  j = arena_alloc (&table->container, sizeof (tab_joined_cell));
  j->hit = 0;
  j->x1 = x1;
  j->y1 = y1;
  j->x2 = ++x2;
  j->y2 = ++y2;
  
  {
    va_list args;
    
    va_start (args, text);
    text_format (table, opt, text, args, &j->contents);
    va_end (args);
  }
  
  opt |= TAB_JOIN;
  
  {
    a_string *cc = &table->cc[x1 + y1 * table->nc];
    unsigned char *ct = &table->ct[x1 + y1 * table->nc];
    const int ofs = table->nc - (x2 - x1);

    int y;
    
    for (y = y1; y < y2; y++)
      {
	int x;
	
	for (x = x1; x < x2; x++)
	  {
	    cc->l = 0;
	    (cc++)->s = (char *) j;
	    *ct++ = opt;
	  }
	
	cc += ofs;
	ct += ofs;
      }
  }
}

/* Set column widths and row heights with a postfix language. */
void
tab_dim (tab_table *t, ...)
{
  unsigned char temp[1024];
  unsigned char *cp = temp;

#define MAX_LABELS 4
  unsigned char *label[MAX_LABELS];

  va_list args;
  int cmd;

  int height = 0, max_height = 0;

#if DEBUGGING
  {
    int i;

    for (i = 0; i < MAX_LABELS; i++)
      label[i] = NULL;
  }
#endif

  va_start (args, t);
  for (cmd = va_arg (args, int); cmd != t_end;
       cmd = va_arg (args, int))
    {
      if (cmd >= t_sentinel)
	{
	  /* Implicit t_con. */
	  assert (cmd >= 0 && cmd <= 65535);
	  
	  *cp++ = t_con - t_end;
	  STORE_2 (cp, cmd);
	  cp += 2;
	  height += 1;
	} else {
	  *cp++ = cmd - t_end;
	  switch (cmd)
	    {
	    case t_end:
	      assert (0);

	    case t_ptw:
	      {
		char *arg = va_arg (args, char *);
		int len = strlen (arg);
	    
		STORE_2 (cp, 2 + 5 + len);
		cp += 2;
		*cp++ = 0;
		STORE_2 (cp, 5);
		cp += 2;
		*cp++ = TAG_FONT_BY_POSITION;
		*cp++ = OUTP_F_BI;
		memcpy (cp, arg, len);
		cp += len;
		height += 1;
	      }
	    break;
	  
	    case t_psw:
	    case t_fsw:
	      cp = 1 + stpcpy (cp, va_arg (args, char *));
	      height += 1;
	      break;

	    case t_pcw:
	    case t_fcw:
	    case t_hgt:
	    case t_nr:
	    case t_nc:
	    case t_dup:
	      height += 1;
	      break;
	  
	    case t_add:
	    case t_sub:
	    case t_mul:
	    case t_div:
	    case t_max:
	    case t_min:
	    case t_str:
	    case t_shr:
	      height -= 1;
	      break;
	  
	    case t_sac:
	    case t_sar:
	      height -= 2;
	      break;

	    case t_scr:
	    case t_srr:
	      height -= 3;
	      break;

	    case t_col:
	    case t_row:
	    case t_nah:
	    case t_naw:
	    case t_neg:
	    case t_xch:
	      break;

	    case t_lbl:
	      {
		int index = va_arg (args, int) - 'a';

		assert (index >= 0 && index < MAX_LABELS);
		label[index] = --cp;
	      }
	    break;
	      
	    case t_jnz:
	      {
		int index = va_arg (args, int) - 'a';

		assert (index >= 0 && index < MAX_LABELS);
		assert (label[index]);
		STORE_2 (cp, cp - label[index]);
		cp += 2;
		height--;
	      }
	    break;
	  
	    case t_con:
	      assert (0);
	  
	    default:
	      assert (0);
	    }
	}

      if (height > max_height)
	max_height = height;

      assert (height >= 0);
    }
  va_end (args);

  if (max_height > t->max_stack_height)
    t->max_stack_height = max_height;
  
  if (t->dim == NULL)
    {
      t->dim = arena_malloc (&t->container, cp - temp + 256);
      STORE_2 (&t->dim[0], cp - temp + 256);
      STORE_2 (&t->dim[2], cp - temp);
    }
  else
    {
      int len = LOAD_2 (&t->dim[2]);
      int size = LOAD_2 (&t->dim[0]);
	      
      if (4 + (cp - temp) + len > size)
	{
	  int new_size = 4 + (cp - temp) + len + 256;
	  t->dim = arena_realloc (&t->container, t->dim, new_size);
	  STORE_2 (&t->dim[0], new_size);
	}
      STORE_2 (&t->dim[2], len + (cp - temp));
    }
	    
  memcpy (&t->dim[4 + LOAD_2 (&t->dim[2]) - (cp - temp)], temp, cp - temp);

#if DEBUGGING
  {
    char *names[t_sentinel - t_end + 1] = 
      {
	"end", "psw", "ptw", "fsw", "pcw", "fcw", "hgt", "nr", "nc", "col",
	"row", "nah", "naw", "con", "add", "sub", "mul", "div", "neg", "max",
	"min", "xch", "dup", "lbl", "jnz", "str", "shr", "sac", "sar", "scr",
	"srr", "sentinel",
      };
    
    int size = LOAD_2 (&t->dim[0]);
    int len = LOAD_2 (&t->dim[2]);
    
    printf ("tab_dim: size=%d, length=%d, max_stack_height=%d ",
	    size, len, max_height);
    for (cp = &t->dim[4]; cp < &t->dim[4 + len]; )
      {
	if (*cp == t_con - t_end)
	  {
	    cp++;
	    printf (" %d", LOAD_2 (cp));
	    cp += 2;
	  }
	else if (*cp == t_ptw - t_end)
	  {
	    cp++;
	    printf (" %s(\"%.*s\")", names[cp[-1]], LOAD_2 (cp) - 7, cp + 7);
	    cp += LOAD_2 (cp);
	  }
	else if (*cp == t_psw - t_end || *cp == t_fsw - t_end)
	  {
	    cp++;
	    printf (" %s(\"%s\")", names[cp[-1]], cp);
	    cp += strlen (cp) + 1;
	  }
	else if (*cp == t_jnz - t_end)
	  {
	    cp++;
	    printf (" %s(%d)", names[cp[-1]], LOAD_2 (cp));
	    cp += 2;
	  }
	else
	  {
	    if (*cp <= t_sentinel - t_end)
	      printf (" %s", names[*cp]);
	    else
	      printf (" (%d)", *cp);
	    cp++;
	  }
	fflush (stdout);
      }
    printf ("\n");
  }
#endif
  
  assert (height == 0);
}

/* Sets all the columns to their natural widths, and all the rows to a
   single line in height. */
void
tab_nat_dim (tab_table *t)
{
  tab_dim (t,
	   t->nc, t_lbl, 'a', 1, t_sub, t_dup, t_dup, t_naw, t_xch, t_sac,
	   t_jnz, 'a',
	   t_hgt, 0, t_nr, t_srr,
	   t_end);
}

/* Miscellaneous. */

/* Outputs text BUF as a table with a single cell having cell options
   OPTIONS, which is a combination of the TAB_* and TAT_*
   constants. */
void
tab_output_text (int options, const char *buf, ...)
{
  tab_table *t = tab_create (1, 1, 0);
  
  if (options & TAT_PRINTF)
    {
      va_list args;
      char *temp_buf = local_alloc (4096);
      
      va_start (args, buf);
      nvsprintf (temp_buf, buf, args);
      buf = temp_buf;
      va_end (args);
    }

  tab_text (t, 0, 0, options &~ TAT_PRINTF, buf);
  tab_flags (t, SOMF_NO_TITLE | SOMF_NO_SPACING);
  if (options & TAT_NOWRAP)
    tab_dim (t, 0, t_naw, 0, t_sac, t_hgt, 0, t_sar, t_end);
  else
    tab_dim (t, 0, t_naw, 0, t_sac, 0, t_nah, 0, t_sar, t_end);
  tab_submit (t);

  if (options & TAT_PRINTF)
    local_free (buf);
}

/* Set table flags to FLAGS. */
void
tab_flags (tab_table *t, unsigned flags)
{
  t->flags = flags;
}

/* Easy, type-safe way to submit a tab table to som. */
void
tab_submit (tab_table *t)
{
  som_table s;

  s.class = &tab_table_class;
  s.ext = t;
  som_submit (&s);
  tab_destroy (t);
}

static tab_table *t;
static outp_driver *d;
int tab_hit;

static void evaluate_dimensions (void);
static void sum_columns (void);

/* Set the current table to TABLE. */
static void
tabi_table (som_table *table)
{
  t = table->ext;
}

/* Set the current output device to DRIVER. */
static void
tabi_driver (outp_driver *driver)
{
  d = driver;

  evaluate_dimensions ();
  sum_columns ();
}

/* Return the number of columns and rows in the table into N_COLUMNS
   and N_ROWS, respectively. */
static void
tabi_count (int *n_columns, int *n_rows)
{
  *n_columns = t->nc;
  *n_rows = t->nr;
}

static void tabi_cumulate (int cumtype, int start, int *end, int max, int *actual);

/* Return the horizontal and vertical size of the entire table,
   including headers, for the current output device, into HORIZ and
   VERT. */
static void
tabi_area (int *horiz, int *vert)
{
  {
    int w, c;
    
    for (c = t->l + 1, w = t->wl + t->wr + t->w[t->l].length;
	 c < t->nc - t->r; c++)
      w += t->w[c].length + t->wrv[c];
    *horiz = w;
  }
  
  {
    int h, r;
    for (r = t->t + 1, h = t->ht + t->hb + t->h[t->t].length;
	 r < t->nr - t->b; r++)
      h += t->h[r].length + t->hrh[r];
    *vert = h;
  }
}

/* Return the column style for this table into STYLE. */
static void
tabi_columns (int *style)
{
  *style = t->col_style;
}

/* Return the number of header rows/columns on the left, right, top,
   and bottom sides into HL, HR, HT, and HB, respectively. */
static void
tabi_headers (int *hl, int *hr, int *ht, int *hb)
{
  *hl = t->l;
  *hr = t->r;
  *ht = t->t;
  *hb = t->b;
}

/* Determines the number of rows or columns (including appropriate
   headers), depending on CUMTYPE, that will fit into the space
   specified.  Takes rows/columns starting at index START and attempts
   to fill up available space MAX.  Returns in END the index of the
   last row/column plus one; returns in ACTUAL the actual amount of
   space the selected rows/columns (including appropriate headers)
   filled. */
static void
tabi_cumulate (int cumtype, int start, int *end, int max, int *actual)
{
  int n;
  som_dimension *d;
  int *r;
  int total;
  
  assert (cumtype == SOM_ROWS || cumtype == SOM_COLUMNS);
  if (cumtype == SOM_ROWS)
    {
      assert (start >= 0 && start < t->nr);
      n = t->nr;
      d = &t->h[start];
      r = &t->hrh[start + 1];
      total = t->ht + t->hb;
    } else {
      assert (start >= 0 && start < t->nc);
      n = t->nc;
      d = &t->w[start];
      r = &t->wrv[start + 1];
      total = t->wl + t->wr;
    }
  
  total += (*d++).length;
  if (total > max)
    {
      if (end)
	*end = start;
      if (actual)
	*actual = 0;
      return;
    }
    
  {
    int x;
      
    for (x = start + 1; x < n; x++)
      {
	int amt = (*d++).length + *r++;
	
	total += amt;
	if (total > max)
	  {
	    total -= amt;
	    break;
	  }
      }

    if (end)
      *end = x;
    
    if (actual)
      *actual = total;
  }
}

/* Return flags set for the current table into FLAGS. */
static void
tabi_flags (unsigned *flags)
{
  *flags = t->flags;
}

/* Render title for current table, with major index X and minor index
   Y.  Y may be zero, or X and Y may be zero, but X should be nonzero
   if Y is nonzero. */
static void
tabi_title (int x, int y)
{
  char buf[1024];
  char *cp;

  if (t->flags & SOMF_NO_TITLE)
    return;
  
  cp = spprintf (buf, "%d.%d", table_num, subtable_num);
  if (x && y)
    cp = spprintf (cp, "(%d:%d)", x, y);
  else if (x)
    cp = spprintf (cp, "(%d)", x);
  if (cur_proc)
    cp = spprintf (cp, " %s", cur_proc);
  cp = stpcpy (cp, ".  ");
  if (t->title.s)
    {
      memcpy (cp, t->title.s, t->title.l);
      cp += t->title.l;
    }
  *cp = 0;
  
  {
    outp_text text;

    text.options = OUTP_T_JUST_LEFT | OUTP_T_FANCY | OUTP_T_HORZ | OUTP_T_VERT;
    text.s.s = buf;
    text.s.l = cp - buf;
    text.h = d->width;
    text.v = d->font_height;
    text.x = 0;
    text.y = d->cp_y;
    d->class->text_draw (d, &text);
  }
}

static int render_strip (int x, int y, int r, int c1, int c2, int r1, int r2);

/* Execute BODY for each value of X from A to B exclusive. */
#define UNROLL_LOOP(X, A, B, BODY)		\
	do					\
	  {					\
	    for (X = A; X < B; X++)		\
	      {					\
		BODY				\
	      }					\
	  }					\
	while (0)

/* Execute PREP, then BODY for each specified value of X: A1...A2, B1...B2,
   C1...C2, in each case not including the second value, running PREP
   before each loop. */
#define UNROLL_3_LOOPS(X, A1, A2, B1, B2, C1, C2, BODY)	\
	do						\
	  {						\
	    UNROLL_LOOP (X, A1, A2, BODY);		\
	    UNROLL_LOOP (X, B1, B2, BODY);		\
	    UNROLL_LOOP (X, C1, C2, BODY);		\
	  }						\
	while (0)

/* Draws the table region in rectangle (X1,Y1)-(X2,Y2), where column
   X2 and row Y2 are not included in the rectangle, at the current
   position on the current output device.  Draws headers as well. */
static void
tabi_render (int x1, int y1, int x2, int y2)
{
  int y, r;
  
  tab_hit++;
  y = d->cp_y;
  if (!(t->flags & SOMF_NO_TITLE))
    y += d->font_height;
  UNROLL_3_LOOPS (r, 0, t->t * 2 + 1, y1 * 2 + 1, y2 * 2,
		  (t->nr - t->b) * 2, t->nr * 2 + 1,

		  int x = d->cp_x;
		  x += render_strip (x, y, r, 0, t->l * 2 + 1, y1, y2);
		  x += render_strip (x, y, r, x1 * 2 + 1, x2 * 2, y1, y2);
		  x += render_strip (x, y, r, (t->nc - t->r) * 2,
				     t->nc * 2 + 1, y1, y2);
		  y += (r & 1) ? t->h[r / 2].length : t->hrh[r / 2];
		  );
}

som_table_class tab_table_class =
  {
    tabi_table,
    tabi_driver,
    
    tabi_count,
    tabi_area,
    NULL,
    NULL,
    tabi_columns,
    NULL,
    tabi_headers,
    NULL,
    tabi_cumulate,
    tabi_flags,
    
    NULL,
    NULL,

    tabi_title,
    tabi_render,
  };

/* Evaluate the postfix language specified on tab_dimensions to
   produce the width and height of the tables' rows and columns. */
static void
evaluate_dimensions (void)
{
  som_dimension *stack;
  som_dimension *tos;

  unsigned char *cp, *end;
  
#if GLOBAL_DEBUGGING
  {
    int i;
    
    for (i = 0; i < t->nc; i++)
      t->w[i].length = NOT_INT;
    for (i = 0; i < t->nr; i++)
      t->h[i].length = NOT_INT;
    
    {
      int error = 0;
      
      int r;

      for (r = 0; r < t->nr; r++)
	{
	  int c;
	  
	  for (c = 0; c < t->nc; c++)
	    {
	      if (t->cc[c + r * t->nc].l == -1)
		{
		  printf (_("uninitialized cell (%d,%d) "
			    "in table size (%d,%d)\n"),
			  c, r, t->nc, t->nr);
		  error = 1;
		}
	    }
	}
      if (error)
	assert (0);
    }
  }
#endif

  debug_printf (("evaluate_dimensions:"));

  assert (t->dim);
  tos = stack = local_alloc (t->max_stack_height * sizeof (som_dimension));
  end = &t->dim[4 + LOAD_2(&t->dim[2])];
  for (cp = &t->dim[4]; cp < end; )
    {
#if DEBUGGING
      char *names[t_sentinel - t_end + 1] = 
	{
	  "end", "psw", "ptw", "fsw", "pcw", "fcw", "hgt", "nr", "nc", "col",
	  "row", "nah", "naw", "con", "add", "sub", "mul", "div", "neg", "max",
	  "min", "xch", "dup", "lbl", "jnz", "str", "shr", "sac", "sar", "scr",
	  "srr", "sentinel",
	};

      if (*cp < t_sentinel - t_end + 1)
	printf (" %s", names[*cp]);
      else
	printf (" (%d)", *cp);
#endif
      
      switch (t_end + *cp++)
	{
#define ARG1 tos[-1].length
#define ARG2 tos[-2].length
#define ARG3 tos[-3].length	
	case t_end:
	  assert (0);

	case t_psw:
	  {
	    outp_text text;
	  
	    text.options = OUTP_T_JUST_LEFT;
	    text.s.s = cp;
	    text.s.l = strlen (cp);
	    d->class->text_metrics (d, &text);

	    tos->length = text.h;
	    tos->stretch = 0;
	    tos->shrink = 0;
	    tos++;
	  
	    cp += text.s.l + 1;

	    debug_printf (("(\"%s\")", text.s.s));
	  }
	break;

	case t_ptw:
	  {
	    outp_text text;
	  
	    text.options = OUTP_T_JUST_LEFT | OUTP_T_FANCY;
	    text.s.s = &cp[2];
	    text.s.l = LOAD_2 (cp) - 2;
	    d->class->text_metrics (d, &text);

	    tos->length = text.h;
	    tos->stretch = 0;
	    tos->shrink = 0;
	    tos++;
	  
	    cp += LOAD_2 (cp);

	    debug_printf (("(\"%.*s\")", text.s.l - 5, &text.s.s[5]));
	  }
	break;
	    
	case t_fsw:
	  {
	    int len = strlen (cp);
	  
	    tos->length = len * d->fixed_width;
	    tos->stretch = 0;
	    tos->shrink = 0;
	    tos++;

	    debug_printf (("(\"%s\")", cp));
	    
	    cp += len + 1;
	  }
	break;

	case t_pcw:
	  tos->length = d->prop_em_width;
	  tos->stretch = 0;
	  tos->shrink = 0;
	  tos++;
	
	  break;

	case t_fcw:
	  tos->length = d->fixed_width;
	  tos->stretch = 0;
	  tos->shrink = 0;
	  tos++;
	  break;

	case t_hgt:
	  tos->length = d->font_height;
	  tos->stretch = 0;
	  tos->shrink = 0;
	  tos++;
	  break;

	case t_nr:
	  tos->length = t->nr;
	  tos->stretch = 0;
	  tos->shrink = 0;
	  tos++;
	  break;

	case t_nc:
	  tos->length = t->nc;
	  tos->stretch = 0;
	  tos->shrink = 0;
	  tos++;
	  break;

	case t_col:
	  assert (ARG1 >= 0 && ARG1 < t->nc);
	  assert (t->w[ARG1].length != NOT_INT);
	  tos[-1] = t->w[ARG1];
	  break;

	case t_row:
	  assert (ARG1 >= 0 && ARG1 < t->nr);
	  assert (t->h[ARG1].length != NOT_INT);
	  tos[-1] = t->h[ARG1];
	  break;

	case t_nah:
	  {
	    int c;
	    int r = ARG1;

	    ARG1 = d->font_height;
	    for (c = 0; c < t->nc; c++)
	      {
		outp_text text;
		unsigned char opt = t->ct[c + r * t->nc];

		assert (t->w[c].length != NOT_INT);
		if (opt & TAB_JOIN)
		  continue;
		text.s = t->cc[c + r * t->nc];
		if (text.s.s == NULL)
		  continue;
		text.options = ((opt & TAB_RICH) ? OUTP_T_FANCY : OUTP_T_NONE);
		text.options |= OUTP_T_HORZ | OUTP_T_JUST_LEFT;
		text.h = t->w[c].length;
		d->class->text_metrics (d, &text);

		if (text.v > ARG1)
		  ARG1 = text.v;
	      }
	    
	    debug_printf (("=%d", ARG1));
	  }
	break;

	case t_naw:
	  {
	    int c = ARG1;
	    int r;

	    assert (c >= 0 && c < t->nc);
	    
	    ARG1 = 0;
	    for (r = 0; r < t->nr; r++)
	      {
		outp_text text;
		unsigned char opt = t->ct[c + r * t->nc];
		
		if (opt & TAB_JOIN)
		  continue;
		text.s = t->cc[c + r * t->nc];
		if (text.s.s == NULL)
		  continue;
		text.options = ((opt & TAB_RICH) ? OUTP_T_FANCY : OUTP_T_NONE);
		text.options |= OUTP_T_JUST_LEFT;
		d->class->text_metrics (d, &text);

		if (text.h > ARG1)
		  ARG1 = text.h;
	      }
	    
	    {
	      int max_width = (d->width
			       - 2 * d->horiz_line_width[OUTP_L_SINGLE]);
	      
	      if (ARG1 > max_width)
		ARG1 = max_width;
	    }
	      
	    debug_printf (("=%d", ARG1));
	  }
	break;
		
	case t_con:
	  tos->length = LOAD_2 (cp);
	  tos->stretch = 0;
	  tos->shrink = 0;
	  tos++;

	  debug_printf (("(%d)", LOAD_2 (cp)));
	
	  cp += 2;
	  break;

	case t_add:
	  ARG2 += ARG1;
	  tos--;
	  break;

	case t_sub:
	  ARG2 -= ARG1;
	  tos--;
	  break;

	case t_mul:
	  ARG2 *= ARG1;
	  tos--;
	  break;

	case t_div:
	  assert (ARG1);
	  ARG2 /= ARG1;
	  tos--;
	  break;

	case t_neg:
	  ARG1 = -ARG1;
	  break;

	case t_max:
	  if (ARG1 > ARG2)
	    ARG2 = ARG1;
	  tos--;
	  break;

	case t_min:
	  if (ARG1 < ARG2)
	    ARG1 = ARG2;
	  tos--;
	  break;

	case t_str:
	  tos[-2].stretch = ARG1;
	  tos--;
	  break;

	case t_shr:
	  tos[-2].shrink = ARG1;
	  tos--;
	  break;

	case t_xch:
	  {
	    som_dimension t = tos[-2];
	    tos[-2] = tos[-1];
	    tos[-1] = t;
	  }
	break;

	case t_dup:
	  tos[0] = tos[-1];
	  tos++;
	  break;

	case t_lbl:
	  assert (0);

	case t_jnz:
	  if (ARG1)
	    cp -= LOAD_2 (cp);
	  else
	    {
	      cp += 2;
	      tos--;
	    }
	  break;
      
	case t_sac:
	  assert (ARG1 >= 0 && ARG1 < t->nc);
	  assert (t->w[ARG1].length == NOT_INT);
	  assert (ARG2 >= 0);
	  t->w[ARG1] = tos[-2];
	  tos -= 2;
	  break;

	case t_sar:
	  assert (ARG1 >= 0 && ARG1 < t->nr);
	  assert (t->h[ARG1].length == NOT_INT);
	  assert (ARG2 >= 0);
	  t->h[ARG1] = tos[-2];
	  tos -= 2;
	  break;

	case t_scr:
	  assert (ARG2 >= 0 && ARG2 < t->nc);
	  assert (ARG1 >= ARG2 && ARG1 <= t->nc);
	  assert (ARG3 >= 0);
	  {
	    int c;
	  
	    for (c = ARG2; c < ARG1; c++)
	      t->w[c] = tos[-3];
	    tos -= 3;
	  }
	  break;

	case t_srr:
	  assert (ARG2 >= 0 && ARG2 < t->nr);
	  assert (ARG1 >= ARG2 && ARG1 <= t->nr);
	  assert (ARG3 >= 0);
	  {
	    int r;

	    for (r = ARG2; r < ARG1; r++)
	      t->h[r] = tos[-3];
	    tos -= 3;
	  }
	  break;

	case t_sentinel:
	  assert (0);
	
	default:
	  assert (0);
	}

#if 0
      debug_printf ((",%d", tos - stack));
#endif
      assert (tos >= stack);
    }

  assert (tos == stack);
  debug_printf (("\n"));
  
  local_free (stack);

#if GLOBAL_DEBUGGING
  {
    int error = 0;
    int i;
    
    for (i = 0; i < t->nc; i++)
      if (t->w[i].length == NOT_INT)
	{
	  printf ("Width of column %d not initialized.\n", i);
	  error = 1;
	}
    for (i = 0; i < t->nr; i++)
      if (t->h[i].length == NOT_INT)
	{
	  printf ("Height of row %d not initialized.\n", i);
	  error = 1;
	}
    if (error)
      assert (0);
  }
#endif
  
#if 0
  {
    int i;
    
    printf ("row heights:");
    for (i = 0; i < t->nr; i++)
      printf (" %d", t->h[i].length);
    printf ("\ncolumn widths:");
    for (i = 0; i < t->nc; i++)
      printf (" %d", t->w[i].length);
    printf ("\n");
  }
#endif
}

/* Figure out rule and header sizes. */
static void  
sum_columns (void)
{
  int i;

  /* Figure out sizes of rules. */
  for (i = 0; i <= t->nr; i++)
    t->hrh[i] = d->horiz_line_spacing[t->trh[i]];
  for (i = 0; i <= t->nc; i++)
    t->wrv[i] = d->vert_line_spacing[t->trv[i]];

  /* Add up header sizes. */
  for (i = 0, t->wl = t->wrv[0]; i < t->l; i++)
    t->wl += t->w[i].length + t->wrv[i + 1];
  for (i = 0, t->ht = t->hrh[0]; i < t->t; i++)
    t->ht += t->h[i].length + t->hrh[i + 1];
  for (i = t->nc - t->r, t->wr = t->wrv[i]; i < t->nc; i++)
    t->wr += t->w[i].length + t->wrv[i + 1];
  for (i = t->nr - t->b, t->hb = t->hrh[i]; i < t->nr; i++)
    t->hb += t->h[i].length + t->hrh[i + 1];

  /* Title. */
  if (!(t->flags & SOMF_NO_TITLE))
    t->ht += d->font_height;
}

/* Render contiguous strip consisting of columns C1...C2, exclusive,
   on row R, at location (X,Y).  Return width of the strip thus
   rendered.

   Renders joined cells, even those outside the strip, within the
   rendering region (C1,R1)-(C2,R2).

   For the purposes of counting rows and columns in this function
   only, horizontal rules are considered rows and vertical rules are
   considered columns. */
static int
render_strip (int x, int y, int r, int c1, int c2, int r1, int r2)
{
  int x_origin = x;

  /* Horizontal rules. */
  if ((r & 1) == 0)
    {
      int hrh = t->hrh[r / 2];
      int c;

      for (c = c1; c < c2; c++)
	{
	  if (c & 1)
	    {
	      int style = t->rh[(c / 2) + (r / 2 * t->nc)];

	      if (style != TAL_0)
		{
		  color clr = {0, 0, 0, 0};
		  rect rct;

		  rct.x1 = x;
		  rct.y1 = y;
		  rct.x2 = x + t->w[c / 2].length;
		  rct.y2 = y + hrh;
		  d->class->line_horz (d, &rct, &clr, style);
		}
	      x += t->w[c / 2].length;
	    } else {
	      color clr = {0, 0, 0, 0};
	      rect rct;
	      outp_styles s;

	      rct.x1 = x;
	      rct.y1 = y;
	      rct.x2 = x + t->wrv[c / 2];
	      rct.y2 = y + hrh;

	      s.t = r > 0 ? t->rv[(c / 2) + (t->nc + 1) * (r / 2 - 1)] : 0;
	      s.b = r < 2 * t->nr ? t->rv[(c / 2) + (t->nc + 1) * (r / 2)] : 0;
	      s.l = c > 0 ? t->rh[(c / 2 - 1) + (t->nc) * (r / 2)] : 0;
	      s.r = c < 2 * t->nc ? t->rh[(c / 2) + (t->nc) * (r / 2)] : 0;

	      if (s.t | s.b | s.l | s.r)
		d->class->line_intersection (d, &rct, &clr, &s);
	      
	      x += t->wrv[c / 2];
	    }
	}
    } else {
      int c;

      for (c = c1; c < c2; c++)
	{
	  if (c & 1)
	    {
	      const int index = (c / 2) + (r / 2 * t->nc);
	      
	      if (!(t->ct[index] & TAB_JOIN))
		{
		  outp_text text;

		  text.options = ((t->ct[index]
				   & (OUTP_T_JUST_MASK | OUTP_T_FANCY))
				  | OUTP_T_HORZ | OUTP_T_VERT);
		  text.s = t->cc[index];
		  if (text.s.s != NULL)
		    {
		      text.h = t->w[c / 2].length;
		      text.v = t->h[r / 2].length;
		      text.x = x;
		      text.y = y;
		      d->class->text_draw (d, &text);
		    }
		} else {
		  tab_joined_cell *j = (tab_joined_cell *) t->cc[index].s;

		  if (j->hit != tab_hit)
		    {
		      j->hit = tab_hit;

		      if (j->x1 == c / 2 && j->y1 == r / 2
			  && j->x2 <= c2 && j->y2 <= r2)
			{
			  outp_text text;

			  text.options = ((t->ct[index]
					   & (OUTP_T_JUST_MASK | OUTP_T_FANCY))
					  | OUTP_T_HORZ | OUTP_T_VERT);
			  text.s = j->contents;
			  text.x = x;
			  text.y = y;
			  
			  {
			    int c;

			    for (c = j->x1, text.h = -t->wrv[j->x2];
				 c < j->x2; c++)
			      text.h += t->w[c].length + t->wrv[c + 1];
			  }
			  
			  {
			    int r;

			    for (r = j->y1, text.v = -t->hrh[j->y2];
				 r < j->y2; r++)
			      text.v += t->h[r].length + t->hrh[r + 1];
			  }
			  d->class->text_draw (d, &text);
			}
		    }
		}
	      x += t->w[c / 2].length;
	    } else {
	      int style = t->rv[(c / 2) + (r / 2 * (t->nc + 1))];

	      if (style != TAL_0)
		{
		  color clr = {0, 0, 0, 0};
		  rect rct;

		  rct.x1 = x;
		  rct.y1 = y;
		  rct.x2 = x + t->wrv[c / 2];
		  rct.y2 = y + t->h[r / 2].length;
		  d->class->line_vert (d, &rct, &clr, style);
		}
	      x += t->wrv[c / 2];
	    }
	}
    }

  return x - x_origin;
}

